"""Functions related to FVA."""
from cobra.flux_analysis import flux_variability_analysis
import logging
import time
log = logging.getLogger(__name__)


def run_fva(model, rxnsOfInterest, proc, fraction_opt):
    log.info("Starting FVA...")
    start_time = time.time()
    s = flux_variability_analysis(model, reaction_list=rxnsOfInterest, fraction_of_optimum=fraction_opt,
                                  processes=proc)
    elapsed_time = time.time() - start_time
    log.info("Total elapsed FVA time: {:.2f} sec".format(elapsed_time))
    return s


def calculate_score(fva_wt_mut_merged):
    rxns_of_interest = fva_wt_mut_merged.index
    # Format the flux ranges to be more usable:
    WTint = {}
    mutantint = {}
    for rn in rxns_of_interest:
        WTint[rn] = [round(fva_wt_mut_merged.loc[rn]["minWT"], 3), round(fva_wt_mut_merged.loc[rn]["maxWT"], 3)]
        mutantint[rn] = [round(fva_wt_mut_merged.loc[rn]["minKO"], 3), round(fva_wt_mut_merged.loc[rn]["maxKO"], 3)]

    # Calculate a score for each pair of flux ranges (WT and mutant):
    score = []  # Score to be compared to the score threshold
    flux_change = {}  # Flux change direction
    for r in rxns_of_interest:  # For the one KO or for each reaction in the group of KOs
        lb = [WTint[r][0], mutantint[r][0]]  # Store the two lower bounds (WT and mutant)
        ub = [WTint[r][1], mutantint[r][1]]  # Store the two upper bounds (WT and mutant)
        if lb == [0, 0]:
            change_lower_bound = 0
        else:
            # Calculate the difference between the lower bounds divided by the biggest absolute lower
            # bound Ex: WT = [-20, 50] mutant = [1, 30] change_lower_bound = abs(1 - 20) / 20 = 0.95
            # change_upper_bound = abs(50 - 30) / 50 = 0.4 ==> score will be 0.95
            change_lower_bound = abs(max(lb) - min(lb)) / max([abs(el) for el in lb])
        if ub == [0, 0]:
            change_upper_bound = 0
        else:
            change_upper_bound = abs(max(ub) - min(ub)) / max([abs(el) for el in ub])
        score.append(max(change_lower_bound, change_upper_bound))  # Choose the max change as the score

        # Determine direction of change
        # If both lower bounds are the same, and both upper bounds are the same ==> no change
        if WTint[r][0] == mutantint[r][0] and WTint[r][1] == mutantint[r][1]:
            flux_change[r] = 0
        # If the WT upper bound is lower than the mutant lower bound ==> significant change
        elif WTint[r][1] < mutantint[r][0]:
            flux_change[r] = 1
        elif WTint[r][0] > mutantint[r][1]:
            flux_change[r] = -1
        elif ((WTint[r][0] <= mutantint[r][0])
              and (WTint[r][1] <= mutantint[r][1])
              and (max(abs(WTint[r][0] - mutantint[r][0]),
                       abs(WTint[r][1] - mutantint[r][1])) > 0)):
            flux_change[r] = 1
        elif ((WTint[r][0] >= mutantint[r][0])
              and (WTint[r][1] >= mutantint[r][1])
              and (abs(WTint[r][0] - mutantint[r][0]) > 0
                   or abs(WTint[r][1] - mutantint[r][1]) > 0)):
            flux_change[r] = -1
        else:
            flux_change[r] = 0
    score_dict = dict(zip(rxns_of_interest, score))
    return score_dict, flux_change



"""Functions related to sampling."""
import time
from cobra.sampling import sample
import logging
log = logging.getLogger(__name__)


def sample_time(model, nsamples, processors, thinning):
    log.info("Starting sampling...")
    start_time = time.time()
    s = sample(model, nsamples, processes=processors, thinning=thinning)
    elapsed_time = time.time() - start_time
    log.info("Total elapsed time: {:.2f} sec".format(elapsed_time))
    return s



"""
Model type detection and import.
"""

from cobra import io
from pathlib import Path
import logging
log = logging.getLogger(__name__)


def get_model_type(model_file):
    file_format = Path(model_file).suffix
    log.info("Detected " + file_format + " model.")
    if not (file_format == '.sbml' or file_format == '.xml' or file_format == '.json' or file_format == ".mat"):
        raise ValueError('Please use a json or sbml or mat file as your model.')
    return file_format


def import_model(model_file):
    log.info("Importing model...")
    file_format = get_model_type(model_file)
    if file_format == ".xml":
        model = io.read_sbml_model(model_file)
    elif file_format == '.json':
        model = io.load_json_model(model_file)
    elif file_format == '.mat':
        model = io.load_matlab_model(model_file)
    # TODO: Add checks for ID name formats (_LPAREN or _92_) (read_sbml_model has f_replace)
    # TODO: Add checks for SUBSYSTEM
    # TODO: Add checks for extra boundary reactions
    # TODO: Remove blocked reactions?
    # TODO: Fix GENE ASSOCIATION? Not sure if need genes
    return model



"""Functions for optimising the WT."""
import logging
log = logging.getLogger(__name__)


def wtopt_r2(rxn, model_clean, eps):
    with model_clean:  # Keeps the modifications to within the with
        # change the objective to the reaction:
        model_clean.objective = rxn
        # optimize for that objective
        FBA = model_clean.optimize(objective_sense="maximize")
        FBA_sol = FBA.objective_value
        log.info("FBA sol: " + str(FBA_sol))
        # set the new bounds as a percentage of the solution
        if FBA_sol > 0:
            lower_bound = FBA_sol * eps
            upper_bound = rxn.upper_bound
        else:
            with model_clean:
                model_clean.objective = rxn
                FBA = model_clean.optimize(objective_sense="minimize")
                FBA_sol = FBA.objective_value
            # print("Min: " + str(FBA_sol))
            lower_bound = FBA_sol * eps
            upper_bound = 1000  # redundant? questionable? what about backwards reactions?
    return lower_bound, upper_bound


def optimise_wt(model, biomass):
    log.info("Objective: "+str(model.objective))
    fbag = model.optimize()  # The optimized objective value
    # The minimum objective value we want to reach:
    obj = biomass * fbag.objective_value
    log.info("Final obj value: "+str(biomass)+"*"+str(fbag.objective_value)+"="+str(obj))
    return obj



"""Functions for exporting and writing files."""
import os
from ..setup.prepare_reactions import parse_rxns
import time


def write_sampling(s_results, out_path, model_name, n_samples, type, fname):
    final_path = str(out_path) + str(fname) + "_" + str(os.path.basename(os.path.splitext(model_name)[0])) + "_sampling_" + str(
                n_samples) + "_" + type + ".csv.gz"
    s_results.to_csv(final_path, index=False, compression='gzip')
    print("Wrote to "+final_path)


def write_fva(fva_results, out_path, model_name, n_samples, type, fname):
    final_path = str(out_path) + str(fname) + "_" + str(os.path.basename(os.path.splitext(model_name)[0])) + "_FVA_" + str(
                n_samples) + "_" + type + ".csv.gz"
    fva_results.to_csv(final_path, index=True, compression='gzip')
    print("Wrote to "+final_path)


def extract_results(s, results, model):
    # Extract from results
    col_names_start = time.time()
    if results == "exchanges":
        if model.exchanges is not None:
            results_columns = [rxn.id for rxn in model.exchanges]
        else:
            results_columns = [col for col in s if col.startswith('EX_')]
    elif results == "all":
        results_columns = [rxn.id for rxn in model.reactions]
    else:  # File containing reaction IDs
        results_columns = parse_rxns(results)
    print("Column names extraction time: " + str(time.time() - col_names_start))
    s_results_round_start = time.time()
    s_results = s[results_columns].round(3)
    print("s_results round time: " + str(time.time() - s_results_round_start))
    return s_results


def extract_fva_results(s, results, model):
    # Extract from results
    if results is None:
        if model.exchanges is not None:
            results_rows = [rxn.id for rxn in model.exchanges]
        else:
            results_rows = [col for col in s if col.startswith('EX_')]
    else:
        results_rows = parse_rxns(results)
    s_results = s.loc[results_rows]
    return s_results


def export_metab_dict(model):
    # Create a metabolite exchange reaction ID to name dict for plotting in R
    metab_dict = {}
    # ex_rxn = model.exchanges.EX_A
    for ex_rxn in model.exchanges:
        name = next(iter(ex_rxn.metabolites)).name
        if len(name) < 30:
            if name != "":
                metab_dict[ex_rxn.id] = name
            else:
                metab_dict[ex_rxn.id] = next(iter(ex_rxn.metabolites)).id
        else:
            if next(iter(ex_rxn.metabolites)).name.endswith("(e)"):
                metab_dict[ex_rxn.id] = next(iter(ex_rxn.metabolites)).id[:-3]
            else:
                metab_dict[ex_rxn.id] = next(iter(ex_rxn.metabolites)).id
    return metab_dict


def export_metab_id_dict(model):
    # Create an exchange reaction ID to metabolite ID dict
    metab_id_dict = {}
    # ex_rxn = model.exchanges.EX_A
    for ex_rxn in model.exchanges:
        metab_id_dict[ex_rxn.id] = next(iter(ex_rxn.metabolites)).id
    return metab_id_dict


def export_gene_to_rxn_dict(model, ids_to_ko):
    # Create a gene to rxn ID using GPRs
    gene_to_rxn_dict = {}
    for gene in ids_to_ko:
        # gene is a list of 1 item for compatibility with using multiple reaction names in one row/group
        rxns = [rxn.id for rxn in model.genes.get_by_id(gene[0]).reactions]
        gene_to_rxn_dict[gene[0]] = " ".join(rxns)
    return gene_to_rxn_dict



"""
Functions for preparing reactions in a metabolic model.
"""


def set_exchanges_rxn_bounds(model, exchange_min):
    """

    :param model:
    :param exchange_min:
    :return: model:
    """
    # Set the exchange reaction bounds
    for rxn in model.exchanges:
        rxn.lower_bound = -exchange_min
        rxn.upper_bound = 1000
    return model


def parse_rxns(filename, sep=" "):
    if filename is None:
        return None
    with open(filename, 'r') as file:
        ids = [i.strip().split(sep) for i in file]
    return ids

